# -*- coding: utf-8 -*-

"""
Auto-compress and decompress data in DynamoDB
"""

import gzip
import json
from typing import Optional
from pynamodb.attributes import BinaryAttribute


class CompressedJSONAttribute(BinaryAttribute):
    """
    A compressed unicode Attribute.

    Dump to unicode json string, encode to unicode bytes string
    and compress it internally.
    """

    def serialize(self, value) -> Optional[bytes]:
        return gzip.compress(
            json.dumps(value, ensure_ascii=False).encode("utf-8")
        )

    def deserialize(self, value):
        return json.loads(gzip.decompress(value).decode("utf-8"), strict=False)


class CompressedUnicodeAttribute(BinaryAttribute):
    """
    A compressed unicode Attribute.

    Encode to unicode bytes string and compress it internally.
    """

    def serialize(self, value: str) -> Optional[bytes]:
        return gzip.compress(value.encode("utf-8"))

    def deserialize(self, value: bytes) -> Optional[str]:
        return gzip.decompress(value).decode("utf-8")


class CompressedBinaryAttribute(BinaryAttribute):
    """
    A compressed binary Attribute.

    Compress it internally.
    """
    def serialize(self, value: bytes) -> Optional[bytes]:
        return gzip.compress(value)

    def deserialize(self, value):
        return gzip.decompress(value)



# -*- coding: utf-8 -*-

"""
Implements the DynamoDB status tracking pattern.
"""

import typing as T
import enum
import uuid
import inspect
import traceback
import dataclasses
from contextlib import contextmanager

from datetime import datetime, timezone

from pynamodb.attributes import (
    UnicodeAttribute,
    NumberAttribute,
    UTCDateTimeAttribute,
    JSONAttribute,
)
from pynamodb.indexes import (
    GlobalSecondaryIndex,
    KeysOnlyProjection,
)
from pynamodb.settings import OperationSettings

from ...models import Model
from ...compat import cached_property


class BaseStatusEnum(int, enum.Enum):
    """
    Base enum class to define the status code you want the tracker to track.

    The value of the status should be an integer. For example, you have a task
    that has the following status:

    .. code-block:: python

        class StatusEnum(BaseStatusEnum):
            s00_todo = 0 # the task is defined but not started
            s03_in_progress = 3 # the task is in progress
            s06_failed = 6 # the task failed
            s09_success = 9 # the task succeeded
            s10_ignore = 10 # the task already failed multiple times, it is ignored

    """

    @property
    def status_name(self) -> str:
        """
        Human readable status name.
        """
        return self.name

    @classmethod
    def from_value(cls, value: int) -> "BaseStatusEnum":
        """
        Get status enum object from it's value.
        """
        return cls(value)

    @classmethod
    def value_to_name(cls, value: int) -> str:
        """
        Convert status value to status name.
        """
        return cls.from_value(value).name

    @classmethod
    def values(cls) -> T.List[int]:
        """
        Return list of all available status values.
        """
        return [status.value for status in cls]


ZERO_PADDING = 4


EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


def utc_now() -> datetime:
    """
    Get time aware utc now datetime object.
    """
    return datetime.utcnow().replace(tzinfo=timezone.utc)


class StatusAndTaskIdIndex(GlobalSecondaryIndex):
    """
    GSI for query by job_id and status.
    """

    class Meta:
        index_name = "status_and_task_id-index"
        projection = KeysOnlyProjection()

    value: T.Union[str, UnicodeAttribute] = UnicodeAttribute(hash_key=True)
    key: T.Union[str, UnicodeAttribute] = UnicodeAttribute(range_key=True)


class TaskLockedError(Exception):
    """
    Raised when a task worker is trying to work on a locked task.
    """

    pass


class TaskIgnoredError(Exception):
    """
    Raised when a task is already in "ignore" status (You need to define).
    """

    pass


# update context manager in-memory cache
_update_context: T.Dict[
    str,  # the Dynamodb item hash key
    T.Dict[
        str,  # item attribute name
        T.Dict[
            str,  # there are three key, "old" (value), "new" (value), "act" (update action)
            T.Any,
        ],
    ],
] = dict()


@dataclasses.dataclass
class BaseDataClass:
    """
    Base dataclass for data and errors.
    """

    @classmethod
    def from_dict(cls, dct: dict):
        return cls(**dct)

    def to_dict(self) -> dict:
        return dataclasses.asdict(self)


@dataclasses.dataclass
class BaseData(BaseDataClass):
    """
    Base dataclass for data attribute, if you want to use a class instead of
    dict to manage the data attribute. You can inherit from this class and
    define your own data fields.
    """

    pass


@dataclasses.dataclass
class BaseErrors(BaseDataClass):
    error: T.Optional[str] = dataclasses.field(default=None)
    traceback: T.Optional[str] = dataclasses.field(default=None)


class BaseStatusTracker(Model):
    """
    The DynamoDB ORM model for the status tracking. You can use one
    DynamoDB table for multiple status tracking jobs.

    TODO: explain what is the problem this pattern is trying to solve.

    Concepts:

    - job: a high-level description of a job, the similar task on different
        items will be grouped into one job. ``job_id`` is the unique identifier
        for a job.
    - task: a specific task on a specific item. ``task_id`` is the unique identifier
        for a task.
    - status: an integer value to indicate the status of a task. The closer to
        the end, the value should be larger, so we can compare the values.

    Attributes:

    :param key: The unique identifier of the task. It is a compound key of
        job_id and task_id. The format is ``{job_id}{separator}{task_id}``
    :param value: Indicate the status of the task. The format is
        ``{job_id}{separator}{status_code}``.
    :param update_time: when the task status is updated
    :param retry: how many times the task has been retried
    :param lock: a concurrency control mechanism. It is an uuid string.
    :param lock_time: when this task is locked.
    :param data: arbitrary data in python dictionary.
    :param errors: arbitrary error data in python dictionary.
    """

    key: T.Union[str, UnicodeAttribute] = UnicodeAttribute(hash_key=True)
    value: T.Union[str, UnicodeAttribute] = UnicodeAttribute()
    update_time: T.Union[datetime, UTCDateTimeAttribute] = UTCDateTimeAttribute(
        default=utc_now,
    )
    retry: T.Union[int, NumberAttribute] = NumberAttribute(default=0)
    lock: T.Union[T.Optional[str], UnicodeAttribute] = UnicodeAttribute(
        default=None,
        null=True,
    )
    lock_time: T.Union[datetime, UTCDateTimeAttribute] = UTCDateTimeAttribute(
        default=EPOCH,
    )
    data: T.Optional[T.Union[dict, JSONAttribute]] = JSONAttribute(
        default=lambda: dict(),
        null=True,
    )
    errors: T.Optional[T.Union[dict, JSONAttribute]] = JSONAttribute(
        default=lambda: dict(),
        null=True,
    )

    _status_and_task_id_index: T.Optional[StatusAndTaskIdIndex] = None

    # one DynamoDB table can serve multiple jobs
    # if you defined a default job id for the table
    # you don't need to explicitly specify the job id in many API
    JOB_ID: T.Optional[str] = None
    # the separator string between job_id and task_id
    SEP = "____"
    # how many digits the max status code have, this ensures that the
    # status can be used in comparison
    STATUS_ZERO_PAD: int = 3
    # how many retry is allowed before we ignore it
    MAX_RETRY: int = 3
    # how long the lock will expire
    LOCK_EXPIRE_SECONDS: int = 900
    # the default status code, means "to do", usually start from 0
    DEFAULT_STATUS: int = 0
    # the status enum class for this tracker
    STATUS_ENUM: T.Optional[T.Type[BaseStatusEnum]] = None

    @classmethod
    def make_key(
        cls,
        task_id: str,
        job_id: T.Optional[str] = None,
    ) -> str:
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        return f"{JOB_ID}{cls.SEP}{task_id}"

    @classmethod
    def make_value(
        cls,
        status: int,
        job_id: T.Optional[str] = None,
    ) -> str:
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        return f"{JOB_ID}{cls.SEP}{str(status).zfill(cls.STATUS_ZERO_PAD)}"

    @cached_property
    def job_id(self) -> str:
        return self.key.split(self.SEP)[0]

    @cached_property
    def task_id(self) -> str:
        """
        Return the task_id part of the key. It should be unique with in a job.
        """
        return self.key.split(self.SEP)[1]

    @property
    def status(self) -> int:
        """
        Return the status value of the task.
        """
        return int(self.value.split(self.SEP)[1])

    @property
    def status_name(self) -> str:
        """
        Return the status name of the task.
        """
        if self.STATUS_ENUM is None:
            return str(self.status)
        else:
            return self.STATUS_ENUM.value_to_name(self.status)

    @classmethod
    def get_one_or_none(
        cls,
        task_id: str,
        consistent_read: bool = False,
        attributes_to_get: T.Optional[T.Sequence[str]] = None,
        settings: OperationSettings = OperationSettings.default,
        job_id: T.Optional[str] = None,
    ) -> T.Optional["BaseStatusTracker"]:
        """
        Get one item by task_id. If the item does not exist, return None.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        return super().get_one_or_none(
            hash_key=cls.make_key(task_id, JOB_ID),
            consistent_read=consistent_read,
            attributes_to_get=attributes_to_get,
            settings=settings,
        )

    def is_item_exists(self) -> bool:
        """
        Check if this Dynamodb item exists.
        """
        return (
            self.get_one_or_none(
                task_id=self.task_id,
                attributes_to_get=[
                    "key",
                ],
            )
            is not None
        )

    @classmethod
    def make(
        cls,
        task_id: str,
        status: int,
        data: T.Optional[dict] = None,
        job_id: T.Optional[str] = None,
    ) -> "BaseStatusTracker":
        """
        A factory method to create new instance of a tracker.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        kwargs = dict(
            key=cls.make_key(task_id, JOB_ID),
            value=cls.make_value(status),
        )
        if data is not None:
            kwargs["data"] = data
        return cls(**kwargs)

    @classmethod
    def new(
        cls,
        task_id: str,
        data: T.Optional[dict] = None,
        save: bool = True,
        job_id: T.Optional[str] = None,
    ) -> "BaseStatusTracker":
        """
        A factory method to create new task with the default status (usually 0).

        :param save: if True, also save the item to dynamodb.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        obj = cls.make(
            task_id=task_id,
            status=cls.DEFAULT_STATUS,
            data=data,
            job_id=JOB_ID,
        )
        if save:
            obj.save()
        return obj

    @classmethod
    def delete_all_by_job_id(
        cls,
        job_id: T.Optional[str] = None,
    ) -> int:
        """
        Delete all item belongs to specific job.

        Note: this method is expensive, it will scan a lot of items.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        ith = 0
        with cls.batch_write() as batch:
            for ith, item in enumerate(
                cls.scan(
                    filter_condition=cls.key.startswith(JOB_ID),
                    attributes_to_get=[
                        "key",
                    ],
                ),
                start=1,
            ):
                batch.delete(item)
        return ith

    @classmethod
    def count_items_by_job_id(
        cls,
        job_id: T.Optional[str] = None,
    ):
        """
        Count number of items belong to specific job.

        Note: this method is expensive, it will scan a lot of items.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        ith = 0
        for ith, item in enumerate(
            cls.scan(
                filter_condition=cls.key.startswith(JOB_ID),
                attributes_to_get=[
                    "key",
                ],
            ),
            start=1,
        ):
            pass
        return ith

    def is_locked(self) -> bool:
        """
        Check if the task is locked.
        """
        if self.lock is None:
            return False
        else:
            now = utc_now()
            return (now - self.lock_time).total_seconds() < self.LOCK_EXPIRE_SECONDS

    def _setup_update_context(self):
        _update_context[self.key] = dict()

    def _rollback_update_context(self):
        for attr_name, dct in _update_context[self.key].items():
            setattr(self, attr_name, dct["old"])

    def _flush_update_context(self):
        actions = [dct["act"] for dct in _update_context[self.key].values()]
        if len(actions):
            # print("flushing update data to Dyanmodb")
            # print(f"actions: {actions}")
            # pynamodb update API will apply the updated data to the current
            # item object.
            self.update(actions=actions)

    def _teardown_update_context(self):
        del _update_context[self.key]

    @contextmanager
    def update_context(self) -> "BaseStatusTracker":
        """
        A context manager to update the attributes of the task.
        If the update fails, the attributes will be rolled back to the original
        value.

        Usage::

            tracker = Tracker.new(job_id="my-job", task_id="task-1")
            with tracker.update_context():
                tracker.set_status(StatusEnum.s03_in_progress)
                tracker.set_data({"foo": "bar"})
        """
        try:
            self._setup_update_context()
            yield self
            self._flush_update_context()
        except Exception as e:
            self._rollback_update_context()
            raise e
        finally:
            self._teardown_update_context()

    def set_status(self, status: int) -> "BaseStatusTracker":
        """
        Set the status of the task. Don't do this directly::

            self.value = self.make_value(self.job_id, ...)
        """
        _update_context[self.key]["value"] = {"old": self.value}
        self.value = self.make_value(status)
        _update_context[self.key]["value"]["new"] = self.value
        _update_context[self.key]["value"]["act"] = BaseStatusTracker.value.set(
            self.value
        )
        return self

    def set_update_time(
        self, update_time: T.Optional[datetime] = None
    ) -> "BaseStatusTracker":
        """
        Set the update time of the task. Don't do this directly::

            self.update_time = ...
        """
        _update_context[self.key]["update_time"] = {"old": self.update_time}
        if update_time is None:
            update_time = utc_now()
        self.update_time = update_time
        _update_context[self.key]["update_time"]["new"] = self.update_time
        _update_context[self.key]["update_time"][
            "act"
        ] = BaseStatusTracker.update_time.set(self.update_time)
        return self

    def set_retry_as_zero(self) -> "BaseStatusTracker":
        """
        Set the retry count to zero. Don't do this directly::

            self.retry = 0
        """
        _update_context[self.key]["retry"] = {"old": self.retry}
        self.retry = 0
        _update_context[self.key]["retry"]["new"] = self.retry
        _update_context[self.key]["retry"]["act"] = BaseStatusTracker.retry.set(
            self.retry
        )
        return self

    def set_retry_plus_one(self) -> "BaseStatusTracker":
        """
        Increase the retry count by one. Don't do this directly::

            self.retry += 1
        """
        _update_context[self.key]["retry"] = {"old": self.retry}
        self.retry += 1
        _update_context[self.key]["retry"]["new"] = self.retry
        _update_context[self.key]["retry"]["act"] = BaseStatusTracker.retry.set(
            BaseStatusTracker.retry + 1
        )
        return self

    def set_locked(self) -> "BaseStatusTracker":
        """
        Set the lock of the task. Don't do this directly::

            self.lock = ...
            self.lock_time = ...
        """
        _update_context[self.key]["lock"] = {"old": self.lock}
        _update_context[self.key]["lock_time"] = {"old": self.lock_time}
        self.lock = uuid.uuid4().hex
        self.lock_time = utc_now()
        _update_context[self.key]["lock"]["new"] = self.lock
        _update_context[self.key]["lock_time"]["new"] = self.lock_time
        _update_context[self.key]["lock"]["act"] = BaseStatusTracker.lock.set(self.lock)
        _update_context[self.key]["lock_time"]["act"] = BaseStatusTracker.lock_time.set(
            self.lock_time
        )
        return self

    def set_unlock(self) -> "BaseStatusTracker":
        """
        Set the lock of the task to None. Don't do this directly::

            self.lock = None
        """
        _update_context[self.key]["lock"] = {"old": self.lock}
        self.lock = None
        _update_context[self.key]["lock"]["new"] = self.lock
        _update_context[self.key]["lock"]["act"] = BaseStatusTracker.lock.set(self.lock)
        return self

    def set_data(self, data: T.Optional[dict]) -> "BaseStatusTracker":
        """
        Logically the data attribute should be mutable,
        make sure don't edit the old data directly
        for example, don't do this::

            self.data["foo"] = "bar"
            self.set_data(self.data)

        Please do this::

            new_data = self.data.copy()
            new_data["foo"] = "bar"
            self.set_data(new_data)
        """
        _update_context[self.key]["data"] = {"old": self.data}
        self.data = data
        _update_context[self.key]["data"]["new"] = self.data
        _update_context[self.key]["data"]["act"] = BaseStatusTracker.data.set(data)
        return self

    def set_errors(self, errors: T.Optional[dict]) -> "BaseStatusTracker":
        """
        Similar to :meth:`Tracker.set_data`. But it is for errors.
        """
        _update_context[self.key]["errors"] = {"old": self.errors}
        self.errors = errors
        _update_context[self.key]["errors"]["new"] = self.errors
        _update_context[self.key]["errors"]["act"] = BaseStatusTracker.errors.set(
            errors
        )
        return self

    def pre_start_hook(self):
        """
        A hook function that will be called before the task is started.
        """
        pass

    def post_start_hook(self):
        """
        A hook function that will be called after the task is finished.
        """
        pass

    @contextmanager
    def start(
        self,
        in_process_status: int,
        failed_status: int,
        success_status: int,
        ignore_status: int,
        debug: bool = False,
    ) -> "BaseStatusTracker":
        """
        A context manager to execute a task, and handle error automatically.

        1. It will set the status to the
        ``in_progress_status`` and set the lock. If the task is already locked,
        it will raise a :class:`TaskLockedError`.
        2. If the task succeeded, it will set the status to the ``success_status``.
        3. If the task fail, it will set the status to the ``failed_status`` and
            log the error to ``.errors`` attribute.
        4. If the task failed N times in a row, it will set the status to the
            ``ignore_status``.
        """
        self.pre_start_hook()

        if debug:
            print(
                "{msg:-^80}".format(
                    msg=(
                        f" ▶️ start task(job_id={self.job_id!r}, "
                        f"task_id={self.task_id!r}, "
                        f"status={self.status_name!r}) "
                    )
                )
            )

        # Handle concurrent lock
        if self.is_locked():
            raise TaskLockedError(f"Task {self.key} is locked.")

        # Handle ignore status
        if self.status == ignore_status:
            raise TaskIgnoredError(
                f"Task {self.key} retry count already exceeded {self.MAX_RETRY}, "
                f"ignore it."
            )

        # mark as in progress
        with self.update_context():
            self.set_status(in_process_status).set_update_time().set_locked()
            if debug:
                print(f"🔓 set status {self.status_name!r} and lock the task.")

        try:
            self._setup_update_context()
            # print("before yield")
            yield self
            if debug:
                print(
                    f"✅ 🔐 task succeeded, "
                    f"set status {self.status_name!r} and unlock the task."
                )
            # print("after yield")
            (
                self.set_status(success_status)
                .set_update_time()
                .set_unlock()
                .set_retry_as_zero()
            )
            # print("end of success logic")
        except Exception as e:  # handling user code
            # print("before error handling")
            # reset the update context
            self._teardown_update_context()
            self._setup_update_context()
            (
                self.set_status(failed_status)
                .set_update_time()
                .set_unlock()
                .set_errors(
                    {
                        "error": repr(e),
                        "traceback": traceback.format_exc(limit=10),
                    }
                )
                .set_retry_plus_one()
            )
            if self.retry >= self.MAX_RETRY:
                self.set_status(ignore_status)
                if debug:
                    print(
                        f"❌ 🔐 task failed {self.MAX_RETRY} times already, "
                        f"set status {self.status_name!r} and unlock the task."
                    )
            else:
                if debug:
                    print(
                        f"❌ 🔐 task failed, "
                        f"set stats {self.status_name!r} and unlock the task."
                    )
            # print("after error handling")
            raise e
        finally:
            # print("before finally")
            self._flush_update_context()
            self._teardown_update_context()

            if debug:
                print(
                    "{msg:-^80}".format(
                        msg=(
                            f" ⏹️ end task(job_id={self.job_id!r}, "
                            f"task_id={self.task_id!r}, "
                            f"status={self.status_name!r}) "
                        )
                    )
                )

            self.post_start_hook()
            # print("after finally")

    @classmethod
    def _get_status_index(cls) -> StatusAndTaskIdIndex:
        """
        Detect the status index object.
        """
        if cls._status_and_task_id_index is None:
            for k, v in inspect.getmembers(cls):
                if isinstance(v, StatusAndTaskIdIndex):
                    cls._status_and_task_id_index = v
                    return cls._status_and_task_id_index
            raise ValueError("you haven't defined a StatusAndTaskIdIndex")
        else:
            return cls._status_and_task_id_index

    @classmethod
    def query_by_status(
        cls,
        status: T.Union[int, T.List[int]],
        limit: int = 10,
        job_id: T.Optional[str] = None,
    ) -> T.Iterable["BaseStatusTracker"]:
        """
        Get task items by status.
        """
        JOB_ID = cls.JOB_ID if job_id is None else job_id
        if isinstance(status, list):
            status_list = status
        else:
            status_list = [status]
        for status in status_list:
            yield from cls._get_status_index().query(
                hash_key=cls.make_value(status, JOB_ID),
                limit=limit,
            )



# -*- coding: utf-8 -*-

"""
Enhance the pynamodb.models.Model class.
"""

import typing as T
import copy as copy_lib

from pynamodb.models import (
    Model as PynamodbModel,
)
from pynamodb.settings import OperationSettings
from pynamodb.exceptions import DeleteError


class ConsoleUrlMaker:
    """
    Create common AWS Dynamodb Console url.
    """

    def table_overview(
        self,
        region_name: str,
        table_name: str,
    ) -> str:
        """
        Create the AWS Console url that can preview Dynamodb Table overview.
        """
        return (
            f"https://{region_name}.console.aws.amazon.com/dynamodbv2/"
            f"home?region={region_name}#"
            f"table?initialTagKey=&name={table_name}&tab=overview"
        )

    def table_items(
        self,
        region_name: str,
        table_name: str,
    ) -> str:
        """
        Create the AWS Console url that can preview items in Dynamodb Table.
        """
        return (
            f"https://{region_name}.console.aws.amazon.com/dynamodbv2/"
            f"home?region={region_name}#item-explorer?"
            f"initialTagKey=&maximize=true&table={table_name}"
        )

    def item_detail(
        self,
        region_name: str,
        table_name: str,
        hash_key: str,
        range_key: T.Optional[str] = None,
    ) -> str:
        """
        Create the AWS Console url that can preview specific Dynamodb item.
        """
        if range_key is None:
            range_key_param = "sk"
        else:
            range_key_param = f"sk={range_key}"
        return (
            f"https://{region_name}.console.aws.amazon.com/dynamodbv2/"
            f"home?region={region_name}#"
            f"edit-item?table={table_name}&itemMode=2&"
            f"pk={hash_key}&"
            f"{range_key_param}&"
            f"ref=%23item-explorer%3Ftable%3D{table_name}&"
            f"route=ROUTE_ITEM_EXPLORER"
        )


console_url_maker = ConsoleUrlMaker()


class Model(PynamodbModel):
    """
    Pynamodb Model with additional features.
    """

    def __init__(
        self,
        hash_key: T.Optional[T.Any] = None,
        range_key: T.Optional[T.Any] = None,
        **attributes,
    ):
        super().__init__(hash_key, range_key, **attributes)
        self.__post_init__()

    def __post_init__(self):
        """
        Allow user to customize the post init behavior.
        For example, it can be used to validate the data.
        """
        pass

    def to_dict(self, copy=False) -> dict:
        """
        Access the item data as a dictionary.
        """
        if copy:
            return copy_lib.deepcopy(self.attribute_values)
        else:
            return self.attribute_values

    @classmethod
    def get_one_or_none(
        cls,
        hash_key: T.Any,
        range_key: T.Optional[T.Any] = None,
        consistent_read: bool = False,
        attributes_to_get: T.Optional[T.Sequence[str]] = None,
        settings: OperationSettings = OperationSettings.default,
    ) -> T.Optional["Model"]:
        """
        Get one Dynamodb item object or None if not exists.
        """
        try:
            return cls.get(
                hash_key=hash_key,
                range_key=range_key,
                consistent_read=consistent_read,
                attributes_to_get=attributes_to_get,
                settings=settings,
            )
        except cls.DoesNotExist:
            return None

    def delete_if_exists(self) -> bool:
        """
        Delete the item if exists. Return True if exists.
        """
        hash_key_attr = self.__class__._hash_key_attribute()
        range_key_attr = self.__class__._range_key_attribute()
        condition = hash_key_attr.exists()
        if range_key_attr:
            condition &= range_key_attr.exists()
        try:
            self.delete(condition=condition)
            return True
        except DeleteError:
            return False

    @classmethod
    def delete_all(cls) -> int:
        """
        Delete all item in a Dynamodb table by scanning all item and delete.
        """
        ith = 0
        with cls.batch_write() as batch:
            for ith, item in enumerate(cls.scan(), start=1):
                batch.delete(item)
        return ith

    @classmethod
    def get_table_overview_console_url(cls) -> str:
        """
        Create the AWS Console url that can preview Dynamodb Table settings.
        """
        return console_url_maker.table_overview(
            region_name=cls.Meta.region,
            table_name=cls.Meta.table_name,
        )

    @classmethod
    def get_table_items_console_url(cls) -> str:
        """
        Create the AWS Console url that can preview items in Dynamodb Table.
        """
        return console_url_maker.table_items(
            region_name=cls.Meta.region,
            table_name=cls.Meta.table_name,
        )

    @property
    def item_detail_console_url(self) -> str:
        """
        Return the AWS Console url that can preview Dynamodb item data.
        """
        klass = self.__class__
        hash_key_name = klass._hash_keyname
        range_key_name = klass._range_keyname
        hash_key_attr = getattr(klass, hash_key_name)
        hash_key_value = getattr(self, hash_key_name)
        kwargs = dict(
            region_name=klass.Meta.region,
            table_name=klass.Meta.table_name,
            hash_key=hash_key_attr.serialize(hash_key_value),
        )
        if range_key_name is not None:
            range_key_attr = getattr(klass, range_key_name)
            range_key_value = getattr(self, range_key_name)
            kwargs["range_key"] = range_key_attr.serialize(range_key_value)
        return console_url_maker.item_detail(**kwargs)



# -*- coding: utf-8 -*-

"""
Store big binary or text in S3 and store the s3 object path in DynamoDB.
"""

import gzip

from pynamodb.attributes import (
    UnicodeAttribute,
)
from ..helpers import sha256, join_s3_uri, split_s3_uri, is_s3_object_exists


class S3BackedAttribute(UnicodeAttribute):
    """
    **中文文档**

    没有办法从 Attribute 对象中访问到上级的 Model 对象, 所以只能自己传入每个 attribute
    的参数.
    """

    bucket_name = None
    key_template = None
    s3_client = None

    _is_s3_backed = True

    def _get_s3_key(self, fingerprint: str) -> str:
        return self.key_template.format(fingerprint=fingerprint)


class S3BackedBigBinaryAttribute(S3BackedAttribute):
    """
    Representation of a binary attribute.
    """
    key_template = "pynamodb-mate/bigbinary/{fingerprint}.dat"

    def serialize(self, value: bytes) -> str:
        fingerprint = sha256(value)
        s3_bucket = self.bucket_name
        s3_key = self._get_s3_key(fingerprint)
        s3_uri = join_s3_uri(s3_bucket, s3_key)
        if not is_s3_object_exists(self.s3_client, s3_bucket, s3_key):
            self.s3_client.put_object(
                Bucket=s3_bucket,
                Key=s3_key,
                Body=gzip.compress(value),
            )
        return s3_uri

    def deserialize(self, value: str) -> bytes:
        s3_bucket, s3_key = split_s3_uri(value)
        res = self.s3_client.get_object(Bucket=s3_bucket, Key=s3_key)
        return gzip.decompress(res["Body"].read())


class S3BackedBigTextAttribute(S3BackedAttribute):
    """
    Representation of a text attribute.
    """
    key_template = "pynamodb-mate/bigtext/{fingerprint}.txt"

    def serialize(self, value: str) -> str:
        b_value = value.encode("utf-8")
        fingerprint = sha256(b_value)
        s3_bucket = self.bucket_name
        s3_key = self._get_s3_key(fingerprint)
        s3_uri = join_s3_uri(s3_bucket, s3_key)
        if not is_s3_object_exists(self.s3_client, s3_bucket, s3_key):
            self.s3_client.put_object(
                Bucket=s3_bucket,
                Key=s3_key,
                Body=gzip.compress(b_value),
            )
        return s3_uri

    def deserialize(self, value: str) -> str:
        s3_bucket, s3_key = split_s3_uri(value)
        res = self.s3_client.get_object(Bucket=s3_bucket, Key=s3_key)
        return gzip.decompress(res["Body"].read()).decode("utf-8")



# -*- coding: utf-8 -*-

"""
implement the abstract cipher.
"""

import json
import hashlib

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from .helpers import bytes_to_base85str, base85str_to_bytes


def str_to_key(s: str) -> bytes:
    m = hashlib.sha256()
    m.update(s.encode("utf-8"))
    return m.digest()


class BaseCipher:
    def encrypt(self, b: bytes) -> bytes: # pragma: no cover
        raise NotImplementedError

    def decrypt(self, b: bytes) -> bytes: # pragma: no cover
        raise NotImplementedError


class AesEcbCipher(BaseCipher):
    """
    A determinative symmetric cipher. same input + same key = same output.
    """

    def __init__(self, key: bytes):
        self._key = key
        self._aes = AES.new(self._key, AES.MODE_ECB)

    def encrypt(self, b: bytes) -> bytes:
        return self._aes.encrypt(pad(b, AES.block_size))

    def decrypt(self, b: bytes) -> bytes:
        return unpad(self._aes.decrypt(b), AES.block_size)


class AesCtrCipher(BaseCipher):
    """
    A non-determinative symmetric cipher. same input + same key != same output.
    """

    def __init__(self, key: bytes):
        self._key = key

    def encrypt(self, b: bytes) -> bytes:
        aes = AES.new(self._key, AES.MODE_CTR)
        token = aes.encrypt(b)
        nonce = aes.nonce
        return json.dumps({
            "nonce": bytes_to_base85str(nonce),
            "token": bytes_to_base85str(token),
        }).encode("ascii")

    def decrypt(self, b: bytes) -> bytes:
        data = json.loads(b.decode("ascii"))
        token = base85str_to_bytes(data["token"])
        nonce = base85str_to_bytes(data["nonce"])
        aes = AES.new(self._key, AES.MODE_CTR, nonce=nonce)
        return aes.decrypt(token)



# -*- coding: utf-8 -*-

import typing
import base64
import hashlib
from botocore.exceptions import ClientError


def bytes_to_base85str(b: bytes) -> str:
    """
    """
    return base64.b85encode(b).decode("utf-8")


def base85str_to_bytes(s: str) -> bytes:
    """
    """
    return base64.b85decode(s.encode("utf-8"))


def sha256(b: bytes) -> str:
    """
    """
    m = hashlib.sha256()
    m.update(b)
    return m.hexdigest()


def split_s3_uri(s3_uri: str) -> typing.Tuple[str, str]:
    """
    Split AWS S3 URI, returns bucket and key.
    """
    parts = s3_uri.split("/")
    bucket = parts[2]
    key = "/".join(parts[3:])
    return bucket, key


def join_s3_uri(bucket: str, key: str) -> str:
    """
    Join AWS S3 URI from bucket and key.
    """
    return "s3://{}/{}".format(bucket, key)


def is_s3_object_exists(s3_client, bucket: str, key: str) -> bool:  # pragma: no cover
    try:
        s3_client.head_object(Bucket=bucket, Key=key)
        return True
    except ClientError as e:
        return False
    except:
        raise


def remove_s3_prefix(s3_client, bucket, prefix):
    res = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix, MaxKeys=1000)
    for content in res.get("Contents", []):
        key = content["Key"]
        s3_client.delete_object(Bucket=bucket, Key=key)



# -*- coding: utf-8 -*-

"""
Use DynamoDB to track the status of many tasks, with error handling,
retry, catch up, etc ...
"""

from .impl import (
    BaseStatusEnum,
    StatusAndTaskIdIndex,
    TaskLockedError,
    TaskIgnoredError,
    BaseData,
    BaseErrors,
    BaseStatusTracker,
)


# -*- coding: utf-8 -*-

"""
Implement Client Side Encryption Attribute for Unicode, Binary, Numbers and
Json data.
"""

import json
import typing
from pynamodb.attributes import BinaryAttribute

from ..cipher import str_to_key, BaseCipher, AesEcbCipher, AesCtrCipher


class SymmetricEncryptedAttribute(object):
    """
    Symmetric Encrypted Attribute
    """
    encryption_key = None  # type: str
    determinative = None  # type: str

    _key = None  # type: bytes

    @property
    def key(self) -> bytes:
        if self._key is None:
            self._key = str_to_key(self.encryption_key)
        return self._key

    _cipher = None  # type: BaseCipher

    @property
    def cipher(self) -> BaseCipher:
        """
        :rtype: SymmtricCipher
        """
        if self._cipher is None:
            if self.determinative is True:
                self._cipher = AesEcbCipher(key=self.key)
            elif self.determinative is False:
                self._cipher = AesCtrCipher(key=self.key)
            else:  # pragma: no cover
                raise ValueError
        return self._cipher


class EncryptedUnicodeAttribute(BinaryAttribute, SymmetricEncryptedAttribute):
    """
    Encrypted Unicode Attribute.
    """
    def serialize(self, value: str) -> bytes:
        return self.cipher.encrypt(value.encode("utf-8"))

    def deserialize(self, value: bytes) -> str:
        return self.cipher.decrypt(value).decode("utf-8")


class EncryptedBinaryAttribute(BinaryAttribute, SymmetricEncryptedAttribute):
    def serialize(self, value: bytes) -> bytes:
        return self.cipher.encrypt(value)

    def deserialize(self, value: bytes) -> bytes:
        return self.cipher.decrypt(value)


class EncryptedNumberAttribute(BinaryAttribute, SymmetricEncryptedAttribute):
    """
    Encrypted Number Attribute.
    """
    def serialize(self, value: typing.Union[int, float]) -> bytes:
        return self.cipher.encrypt(json.dumps(value).encode("ascii"))

    def deserialize(self, value: bytes) -> typing.Union[int, float]:
        return json.loads(self.cipher.decrypt(value).decode("ascii"))


class EncryptedJsonAttribute(EncryptedNumberAttribute):
    """
    Encrypted JSON data Attribute.
    """
    pass



# -*- coding: utf-8 -*-

from ._version import __version__

__short_description__ = "Provide Additional Features for pynamodb."
__github_username__ = "MacHu-GWU"
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = __author__
__maintainer_email__ = __author_email__


try:
    from pynamodb.connection import Connection
    from pynamodb.indexes import (
        GlobalSecondaryIndex,
        LocalSecondaryIndex,
        KeysOnlyProjection,
        IncludeProjection,
        AllProjection,
    )
    from pynamodb.constants import (
        BINARY,
        BINARY_SET,
        BOOLEAN,
        LIST,
        MAP,
        NULL,
        NUMBER,
        NUMBER_SET,
        STRING,
        STRING_SET,
        PAY_PER_REQUEST_BILLING_MODE,
        PROVISIONED_BILLING_MODE,
    )
    from pynamodb.attributes import (
        Attribute as CustomAttribute,
        DiscriminatorAttribute,
        BinaryAttribute,
        BinarySetAttribute,
        UnicodeAttribute,
        UnicodeSetAttribute,
        JSONAttribute,
        BooleanAttribute,
        NumberAttribute,
        NumberSetAttribute,
        VersionAttribute,
        TTLAttribute,
        UTCDateTimeAttribute,
        NullAttribute,
        MapAttribute,
        DynamicMapAttribute,
        ListAttribute,
    )
    from pynamodb.transactions import (
        TransactGet,
        TransactWrite,
    )
    from pynamodb.signals import pre_dynamodb_send, post_dynamodb_send
except ImportError as e:  # pragma: no cover
    print(e)
    pass
except:  # pragma: no cover
    raise


try:
    from .attributes.s3backed import (
        S3BackedBigBinaryAttribute,
        S3BackedBigTextAttribute,
    )
    from .attributes.compressed import (
        CompressedJSONAttribute,
        CompressedUnicodeAttribute,
        CompressedBinaryAttribute,
    )
    from .models import Model
except ImportError as e:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise


try:
    from .attributes.encrypted import (
        EncryptedNumberAttribute,
        EncryptedUnicodeAttribute,
        EncryptedBinaryAttribute,
        EncryptedJsonAttribute,
    )
except ImportError as e:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise


try:
    from . import patterns
except ImportError as e:  # pragma: no cover
    pass
except:  # pragma: no cover
    raise
